# -*- coding: utf-8 -*-
# Copyright (c) 2019 Contributors as noted in the AUTHORS file
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# System imports

# Third-party imports

# Local imports

__all__ = [
    '__title__', '__summary__', '__uri__', '__version__',
    '__author__', '__email__', '__license__', '__copyright__',
]

__title__ = 'lookups'
__summary__ = 'DCI lookups in Python (inspired by Netbeans Platform Lookups API)'
__uri__ = 'https://github.com/AxelVoitier/lookups'

__version__ = '0.2.0'

__author__ = 'Axel Voitier'
__email__ = 'axel.voitier@gmail.com'

__license__ = 'MPL-2.0'
__copyright__ = 'Copyright 2019 Lookups contributors (see AUTHORS.md)'
